/*
 * A simple XML parser that converts interface declarations into
 * declaration in the C language 
 *
 * Written by Mlanie Bats, <melanie.bats@utbm.fr>
 *
 * This program is using libxml2.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <libxml/tree.h>
#include <libxml/xpath.h>

/* A simple function that returns a string converted to upper case */
static void toupper_str(char* dst, const char* src)
{
  for ( ; *src != '\0' ; src++, dst++)
    *dst = toupper(*src);
  *dst = '\0';
}

static void print_code_tag (xmlNodePtr node, const char *cur_domain)
{
  char *code_domain = xmlGetProp(node, "domain");
  
  if (code_domain && strcmp(code_domain, cur_domain))
    return;
  
  printf("/* Begin custom code */\n%s\n/* End custom code */\n",
	 xmlNodeGetContent(node));
}

/* Generate the include file corresponding to the interface described
   in the given xmldoc file. */
static void generate_include(xmlDocPtr xmldoc, int interface_id)
{
  char *interface_name, *INTERFACE_NAME;

  xmlNodePtr interface_node, method_node;
  interface_node = xmldoc->children;

  interface_name = xmlGetProp (interface_node, "name");
  INTERFACE_NAME = alloca(strlen(interface_name)+1);
  toupper_str(INTERFACE_NAME, interface_name);

  /* Prevent multiple inclusion */
  printf("#ifndef __%s_H__\n", INTERFACE_NAME);
  printf("#define __%s_H__\n\n", INTERFACE_NAME);

  /* Needed headers */
  printf("#include <karm/karm.h>\n");
  printf("#include <kos/types.h>\n");
  printf("#include <kos/errno.h>\n\n");
  
  /* Define some macros */
  printf("#define INTERFACE_%s __kos_interface_%s\n",
	 INTERFACE_NAME, interface_name);
  printf("#define INTERFACE_%s_ID     %d\n",
	 INTERFACE_NAME, interface_id);
  printf("#define %s_OPS(ops) (( struct __kos_interface_%s *)(ops))\n\n",
	 INTERFACE_NAME, interface_name);

  /* Print the code included in the <code> tag */
  for (method_node = interface_node->children ; 
       method_node != NULL; 
       method_node = method_node->next)
    {
      if ((method_node->type == XML_ELEMENT_NODE) && 
	  (!strcmp(method_node->name, "code")))
	print_code_tag(method_node, "kernel");
    }

  /* Begin declaration of the interface */
  printf("\n struct __kos_interface_%s {\n", interface_name);
  
  for (method_node = interface_node->children ; 
       method_node != NULL; 
       method_node = method_node->next)
    {
      /* Print the prototype of each method */
      if ((method_node->type == XML_ELEMENT_NODE) && 
	  (!strcmp(method_node->name, "method")))
	{
	  xmlNodePtr arg_node;
	  char *method_name = xmlGetProp (method_node , "name");

	  /* Return type is always result_t */
	  printf("   result_t (*%s)(", method_name);

	  /* Print the arguments */
	  for (arg_node = method_node->children ; 
	       arg_node != NULL; 
	       arg_node = arg_node->next)
	    {
	      if (arg_node->type == XML_ELEMENT_NODE)
		{
		  char *arg_type = xmlGetProp(arg_node, "type");
		  char *arg_name = xmlGetProp(arg_node, "name");
		  if ((!arg_type) || (!arg_name))
		    {
		      fprintf(stderr,
			      "Invalid arg in method %s in interface %s\n",
			      method_name, interface_name);
		      exit(EXIT_FAILURE);
		    }

		  printf("%s %s", arg_type, arg_name);
		  if (arg_node->next->next != NULL)
		    {
		      printf(", ");
		    }
		}
	    }
	  
	  /* End of method */
	  printf(");\n");
	}
    }

  /* End of interface */
  printf("};\n\n");

  /* Define a macro coresponding to the interface size */
  printf("#define INTERFACE_%s_NB_OPS INTERFACE_SIZE(struct __kos_interface_%s)\n",
	 INTERFACE_NAME, interface_name);
  printf("#endif\n");  
}

/* Generate the description of an interface (for the interface-desc.c
   file) */
static void generate_description(xmlDocPtr xmldoc)
{
  char *interface_name, *INTERFACE_NAME;
  int number_of_arg;
  
  xmlNodePtr interface_node, method_node, arg_node;
  interface_node = xmldoc->children;

  interface_name = xmlGetProp (interface_node, "name");
  INTERFACE_NAME = alloca(strlen(interface_name)+1);
  toupper_str(INTERFACE_NAME, interface_name);

  /* Beginning of the description */
  printf("[INTERFACE_%s_ID] = { \"%s\" , INTERFACE_%s_NB_OPS, NB_ARGS_ARRAY(",
	 INTERFACE_NAME, interface_name, INTERFACE_NAME);
  
  /* Compute number of arguments for each method */
  for (method_node = interface_node->children ; 
       method_node != NULL ; 
       method_node = method_node->next)
    {
      if ((method_node->type == XML_ELEMENT_NODE)
	  &&(!strcmp(method_node->name, "method")))
	{	      
	  number_of_arg = 0;
	  for(arg_node = method_node->children ; 
	      arg_node != NULL ; 
	      arg_node = arg_node->next)
	    {
	      if (arg_node->type == XML_ELEMENT_NODE)
		{
		  number_of_arg = number_of_arg + 1;
		}
	    }

	  printf("%i", number_of_arg);
	  
	  if (method_node->next->next != NULL)
	    {
	      printf(", ");
	    }
	  
	}
    }

  /* End of interface description */
  printf(") },\n");
}

static void generate_header_stub(xmlDocPtr xmldoc, int interface_id)
{
  char *interface_name, *INTERFACE_NAME;
  char *method_name;
  int id_method = 0;
  
  xmlNodePtr interface_node, method_node;
  interface_node = xmldoc->children;

  interface_name = xmlGetProp (interface_node, "name");
  INTERFACE_NAME = alloca(strlen(interface_name)+1);
  toupper_str(INTERFACE_NAME, interface_name);

  /* Print the code included in the <code> tag */
  for (method_node = interface_node->children ; 
       method_node != NULL ; 
       method_node = method_node->next)
    {
      if ((method_node->type == XML_ELEMENT_NODE) && 
	  (!strcmp(method_node->name, "code")))
	print_code_tag(method_node, "user");
    }
  

  /* Beginning of the description */
  printf("\n#define KOS_INTERFACE_%s_ID %d\n", INTERFACE_NAME, interface_id);

  for (method_node = interface_node->children ; 
       method_node != NULL ; 
       method_node = method_node->next)
    {
      if ((method_node->type == XML_ELEMENT_NODE)
	  &&(!strcmp(method_node->name, "method")))
	{ 
	  int number_of_arg = 0;
	  xmlNodePtr arg_node;
	  char *METHOD_NAME;
	  char *domain;

	  domain = xmlGetProp (method_node, "domain");
	  if (domain && strcmp(domain, "user"))
	    continue;

	  method_name = xmlGetProp (method_node , "name");
	  METHOD_NAME = alloca(strlen(interface_name)+1);
	  toupper_str(METHOD_NAME, method_name);

	  printf("#define KOS_INTERFACE_%s_%s %i\n",
		 INTERFACE_NAME, METHOD_NAME, id_method);
	  id_method = id_method +1;

	  printf("extern result_t __kos_sys_%s_%s (int /* rd */",
		 interface_name, method_name);

	  for(arg_node = method_node->children ; 
	      arg_node != NULL ; 
	      arg_node = arg_node->next)
	    {
	      if (arg_node->type == XML_ELEMENT_NODE)
		{
		  char *arg_type = xmlGetProp(arg_node, "type");
		  char *arg_name = xmlGetProp(arg_node, "name");
		  if ((!arg_type) || (!arg_name))
		    {
		      fprintf(stderr,
			      "Invalid arg #%d of method %s in interface %s\n",
			      number_of_arg+1, method_name, interface_name);
		      exit(EXIT_FAILURE);
		    }
		  
		  number_of_arg = number_of_arg + 1;

		  /* We skip the "struct ures*" first argument, since
		     it corresponds to the "int rd" argument already
		     printed above */
		  if (number_of_arg == 1)
		    continue;
		  
		  printf(", %s %s", arg_type, arg_name);
		} 
	    }
	  printf(");\n");
	}
    }
  printf("\n");

}

static void generate_code_stub(xmlDocPtr xmldoc)
{
  char *interface_name, *INTERFACE_NAME;
  int id_method = 0;
  
  xmlNodePtr interface_node, method_node;
  interface_node = xmldoc->children;

  interface_name = xmlGetProp (interface_node, "name");
  INTERFACE_NAME = alloca(strlen(interface_name)+1);
  toupper_str(INTERFACE_NAME, interface_name);

  for (method_node = interface_node->children ; 
       method_node != NULL ; 
       method_node = method_node->next)
    {
      if ((method_node->type == XML_ELEMENT_NODE)
	  &&(!strcmp(method_node->name, "method")))
	{
	  xmlNodePtr arg_node;
	  char *method_name;
	  int number_of_arg = 0;
	  char *METHOD_NAME;
	  char *domain;

	  domain = xmlGetProp (method_node, "domain");
	  if (domain && strcmp(domain, "user"))
	    continue;

	  method_name = xmlGetProp (method_node , "name");
	  METHOD_NAME = alloca(strlen(interface_name)+1);
	  toupper_str(METHOD_NAME, method_name);

	  printf("result_t __kos_sys_%s_%s (int __rd__",
		 interface_name, method_name);

	  for(arg_node = method_node->children ; 
	      arg_node != NULL ; 
	      arg_node = arg_node->next)
	    {
	      if (arg_node->type == XML_ELEMENT_NODE)
		{
		  char *arg_type = xmlGetProp(arg_node, "type");
		  char *arg_name = xmlGetProp(arg_node, "name");
		  if ((!arg_type) || (!arg_name))
		    {
		      fprintf(stderr,
			      "Invalid arg #%d of method %s in interface %s\n",
			      number_of_arg+1, method_name, interface_name);
		      exit(EXIT_FAILURE);
		    }
		  
		  number_of_arg = number_of_arg + 1;

		  /* We skip the "struct ures*" first argument, since
		     it corresponds to the "int rd" argument already
		     printed above */
		  if (number_of_arg == 1)
		    continue;
		  
		  printf(", %s %s", arg_type, arg_name);
		}
	    }

	  if (number_of_arg <= 0)
	    {
	      fprintf(stderr,
		      "Invalid number of arguments (%d) for method %s in interface %s\n",
		      number_of_arg, method_name, interface_name);
	      exit(EXIT_FAILURE);
	    }

	  printf(")\n{\n");
	  printf("  return __kos_do_syscall%d (__rd__, KOS_INTERFACE_%s_%s",
		 number_of_arg-1, INTERFACE_NAME, METHOD_NAME);
	  
	  number_of_arg = 0;
	  for(arg_node = method_node->children ; 
	      arg_node != NULL ; 
	      arg_node = arg_node->next)
	    {
	      if (arg_node->type == XML_ELEMENT_NODE)
		{
		  number_of_arg = number_of_arg + 1;

		  /* We skip the "struct ures*" first argument, since
		     it corresponds to the "int rd" argument already
		     printed above */
		  if (number_of_arg == 1)
		    continue;

		  printf(",\n\t\t(unsigned long)%s", xmlGetProp(arg_node, "name"));		  
		}
	    }
	  printf(");\n}\n\n");

	  id_method = id_method +1;	  
	}
    }
}

int validate_document(xmlDocPtr doc, char *dtd_file)
{
  xmlDtdPtr dtd;
  xmlValidCtxtPtr ctxt; 
  int result;

  dtd = xmlParseDTD(NULL, dtd_file);
  
  if (!dtd)
    {
      fprintf(stderr, "%s:%d: Problem, unable to load %s\n", 
	      __FILE__, __LINE__, dtd_file);
      exit (EXIT_FAILURE);
    }

  ctxt = xmlNewValidCtxt();
  
  if(!ctxt)
    {
      fprintf(stderr, "%s:%d: Problem, unable to alloc context\n", 
	      __FILE__, __LINE__);
      exit (EXIT_FAILURE);
    }
  
  result = xmlValidateDtd(ctxt, doc, dtd);

  xmlFreeValidCtxt(ctxt);

  /* When xmlValidateDtd returns 0, the document is not valid, when it
     returns 1, the document is valid */
  if(result == 0)
    {
      return -1;
    }

  return 0;
}

int main (int argc, char *argv[])
{
  int i;
  enum { GENERATE_HEADER, GENERATE_DESCRIPTION,
	 GENERATE_HEADER_STUB, GENERATE_CODE_STUB } option;
  char *dtd_file = NULL;
  char *interface_file;
  char *interface_id_string;
  int interface_id;
  xmlDocPtr xmldoc;

  while ((i = getopt (argc, argv, "sdi:l:c:")) != -1)
    switch(i)
      {
      case 'c':
	dtd_file = optarg;
	break;
      case 's':
	option = GENERATE_CODE_STUB;
	break;
      case 'i':
	option = GENERATE_HEADER;
	interface_id_string = optarg;
	interface_id = atoi(interface_id_string);
	break;
      case 'd':
	option = GENERATE_DESCRIPTION;
	break;
      case 'l':
	option = GENERATE_HEADER_STUB;
	interface_id_string = optarg;
	interface_id = atoi(interface_id_string);
	break;
      default:
	printf("Invalid option\n");
      }

  interface_file = argv[optind];

  xmldoc = xmlParseFile (interface_file);

  if (!xmldoc)
    {
      fprintf(stderr, "%s:%d: Problem, unable to load %s\n", 
	      __FILE__, __LINE__, interface_file);
      exit (EXIT_FAILURE);
   }

  if(dtd_file)
    {
      if(validate_document(xmldoc, dtd_file) < 0)
	{
	  fprintf(stderr, "Error while validating %s\n",
		  interface_file);
	  exit (EXIT_FAILURE);
	}
    }

  if (option == GENERATE_HEADER)
    {
      generate_include(xmldoc, interface_id);
    }
  else if(option == GENERATE_DESCRIPTION)
    {
      generate_description(xmldoc);
    }
  else if(option == GENERATE_HEADER_STUB)
    {
      generate_header_stub(xmldoc, interface_id);
    }
  else if (option == GENERATE_CODE_STUB)
    {
      generate_code_stub(xmldoc);
    }

  return 0;
} 
